/************************************************************************/
/* gtc.c - GPL Telescope Control Library Source				*/
/************************************************************************/
/*
 *  (c) 2004 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dlfcn.h>

#include <gtc/gtctypes.h>
#include <gtc/gtc_p.h>
#include <gtc/gtcdef.h>


#define EMPTY_STRING	""

/* GTC error code strings */
const char *GtcErrorString[] = {
	"GTC error base",
	"Hard limit reached",
	"Module loading failed",
	"Module is in use",
	"Alignment failed",
	"Target below horizon",
	"Internal error occurred",
	"Invalid command or argument",
	"Missing required command argument",
	"Device not present",
};


/* library support functions */
int gtc_library_get_version (char **version)
{
	if (version == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*version = VERSION;
	return 0;
}


/* module loading/unloading */
int gtc_module_load (GtcModule **gmp, char *path)
{
	int rc;
	GtcModule *gm = NULL;
	GtcModuleInfo *gmi = NULL;
	GtcVector *gv = NULL;
	GtcLoadFunction load = NULL;

	if (gmp == NULL || path == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*gmp = NULL;
	gm = calloc (1, sizeof (GtcModule));
	if (gm == NULL)
		return GTC_RC_NOMEMORY;
	gmi = calloc (1, sizeof (GtcModuleInfo));
	if (gmi == NULL) {
		free (gm);
		return GTC_RC_NOMEMORY;
	}
	gm->moduleinfo = gmi;
	gv = calloc (1, sizeof (GtcVector));
	if (gv == NULL) {
		free (gmi);
		free (gm);
		return GTC_RC_NOMEMORY;
	}
	gmi->vector = gv;

	gm->module = dlopen (path, RTLD_NOW);
	if (gm->module == NULL) {
		free (gv);
		free (gmi);
		free (gm);
		return GTC_RC_MODULELOADERROR;
	}

	load = dlsym (gm->module, "module_load");
	if (load == NULL) {
		dlclose (gm->module);
		free (gv);
		free (gmi);
		free (gm);
		return GTC_RC_MODULELOADERROR;
	}

	gm->unload = dlsym (gm->module, "module_unload");
	if (gm->unload == NULL) {
		dlclose (gm->module);
		free (gv);
		free (gmi);
		free (gm);
		return GTC_RC_MODULELOADERROR;
	}

	rc = (*load) (gmi);
	if (rc != 0) {
		dlclose (gm->module);
		free (gv);
		free (gmi);
		free (gm);
		return GTC_RC_MODULELOADERROR;
	}
	gm->path = strdup (path);
	if (gm->path == NULL) {
		dlclose (gm->module);
		free (gv);
		free (gmi);
		free (gm);
		return GTC_RC_NOMEMORY;
	}
	*gmp = gm;
	return 0;
}


int gtc_module_get_name (GtcModule *gm, char **name)
{
	if (gm == NULL || name == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*name = NULL;
	if (gm->moduleinfo->name == NULL)
		*name = EMPTY_STRING;
	else
		*name = gm->moduleinfo->name;
	return 0;
}


int gtc_module_get_description (GtcModule *gm, char **description)
{
	if (gm == NULL || description == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*description = NULL;
	if (gm->moduleinfo->description == NULL)
		*description = EMPTY_STRING;
	else
		*description = gm->moduleinfo->description;
	return 0;
}


int gtc_module_get_version (GtcModule *gm, char **version)
{
	if (gm == NULL || version == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*version = NULL;
	if (gm->moduleinfo->version == NULL)
		*version = EMPTY_STRING;
	else
		*version = gm->moduleinfo->version;
	return 0;
}


int gtc_module_get_author (GtcModule *gm, char **author)
{
	if (gm == NULL || author == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*author = NULL;
	if (gm->moduleinfo->author == NULL)
		*author = EMPTY_STRING;
	else
		*author = gm->moduleinfo->author;
	return 0;
}


int gtc_module_get_path (GtcModule *gm, char **path)
{
	if (gm == NULL || path == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*path = NULL;
	if (gm->path == NULL)
		*path = EMPTY_STRING;
	else
		*path = gm->path;
	return 0;
}


int gtc_module_get_usecount (GtcModule *gm, int *usecount)
{
	if (gm == NULL || usecount == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*usecount = gm->refcount;
	return 0;
}


int gtc_module_unload (GtcModule **gmp)
{
	int rc;
	GtcModule *gm = NULL;

	if (gmp == NULL || *gmp == NULL)
		return GTC_RC_INVALIDARGUMENT;
	gm = *gmp;
	if (gm->module == NULL || gm->moduleinfo == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (gm->unload == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (gm->refcount > 0)
		return GTC_RC_MODULEINUSE;

	rc = (*gm->unload) (gm->moduleinfo);
	if (rc != 0)
		return rc;
	dlclose (gm->module);
	free (gm->moduleinfo->vector);
	free (gm->moduleinfo);
	free (gm->path);
	free (gm);
	*gmp = NULL;
	return 0;
}



/* scope connect/disconnect */
int gtc_scope_new (GtcScope **gsp, GtcModule *gm)
{
	GtcScope *gs = NULL;
	if (gsp == NULL || gm == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*gsp = NULL;
	gs = calloc (1, sizeof (GtcScope));
	if (gs == NULL)
		return GTC_RC_NOMEMORY;
	gs->module = gm;
	gm->refcount++;
	*gsp = gs;
	return 0;
}


int gtc_scope_connect (GtcScope *gs, char *device)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (gs->flags & GTC_FLAG_DEVICECONNECTED)
		return GTC_RC_DEVICEBUSY;
	if (gs->module->moduleinfo->vector->scope_connect == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_connect) (gs, device);
	if (rc == 0) {
		gs->flags |= GTC_FLAG_DEVICECONNECTED;
	}
	return rc;
}


int gtc_scope_close (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_close == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_close) (gs);
	if (rc == 0) {
		gs->flags -= (gs->flags & GTC_FLAG_DEVICECONNECTED);
	}
	return rc;
}


int gtc_scope_free (GtcScope **gsp)
{
	GtcScope *gs = NULL;
	if (gsp == NULL || *gsp == NULL)
		return GTC_RC_INVALIDARGUMENT;
	gs = *gsp;
	gs->module->refcount--;
	free (gs);
	*gsp = NULL;
	return 0;
}


/* scope model/firmware information */
int gtc_scope_get_model (GtcScope *gs, char **model)
{
	int rc;
	if (gs == NULL || model == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_model == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_model) (gs, model);
	return rc;
}


int gtc_scope_get_firmware_version (GtcScope *gs, char **version)
{
	int rc;
	if (gs == NULL || version == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_firmware_version == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_firmware_version) (gs,
						version);
	return rc;
}


int gtc_scope_get_firmware_date (GtcScope *gs, char **date)
{
	int rc;
	if (gs == NULL || date == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_firmware_date == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_firmware_date) (gs, date);
	return rc;
}



/* scope date/time */
int gtc_scope_get_datetime (GtcScope *gs, GtcDate *gd, GtcTime *gt)
{
	int rc;
	if (gs == NULL || gd == NULL || gt == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_datetime == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_datetime) (gs, gd, gt);
	return rc;
}


int gtc_scope_set_datetime (GtcScope *gs, GtcDate *gd, GtcTime *gt)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (gd == NULL && gt == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_datetime == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_datetime) (gs, gd, gt);
	return rc;
}


int gtc_scope_get_utc_offset (GtcScope *gs, double *offset)
{
	int rc;
	if (gs == NULL || offset == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_utc_offset == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_utc_offset) (gs, offset);
	return rc;
}


int gtc_scope_set_utc_offset (GtcScope *gs, double offset)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_utc_offset == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_utc_offset) (gs, offset);
	return rc;
}


int gtc_scope_get_sidereal_time (GtcScope *gs, GtcTime *gt)
{
	int rc;
	if (gs == NULL || gt == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_sidereal_time == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_sidereal_time) (gs, gt);
	return rc;
}


int gtc_scope_set_sidereal_time (GtcScope *gs, GtcTime *gt)
{
	int rc;
	if (gs == NULL || gt == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_sidereal_time == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_sidereal_time) (gs, gt);
	return rc;
}



/* scope mount/site/location */
int gtc_scope_autoalign (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_autoalign == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_autoalign) (gs);
	return rc;
}


int gtc_scope_home (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_home == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_home) (gs);
	return rc;
}


int gtc_scope_get_alignment (GtcScope *gs, int *alignment)
{
	int rc;
	if (gs == NULL || alignment == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_alignment == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_alignment) (gs, alignment);
	return rc;
}


int gtc_scope_set_alignment (GtcScope *gs, int alignment)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_alignment == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_alignment) (gs, alignment);
	return rc;
}


int gtc_scope_get_longitude (GtcScope *gs, double *longitude)
{
	int rc;
	if (gs == NULL || longitude == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_longitude == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_longitude) (gs, longitude);
	return rc;
}


int gtc_scope_set_longitude (GtcScope *gs, double longitude)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_longitude == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_longitude) (gs, longitude);
	return rc;
}


int gtc_scope_get_latitude (GtcScope *gs, double *latitude)
{
	int rc;
	if (gs == NULL || latitude == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_latitude == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_latitude) (gs, latitude);
	return rc;
}


int gtc_scope_set_latitude (GtcScope *gs, double latitude)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_latitude == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_latitude) (gs, latitude);
	return rc;
}


int gtc_scope_get_current_site (GtcScope *gs, int *site)
{
	int rc;
	if (gs == NULL|| site == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_current_site == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_current_site) (gs, site);
	return rc;
}


int gtc_scope_set_current_site (GtcScope *gs, int site)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_current_site == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_current_site) (gs, site);
	return rc;
}



/* scope RA/DEC motion */
int gtc_scope_slew_east_start (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_east_start == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_east_start) (gs);
	return rc;
}


int gtc_scope_slew_east_stop (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_east_stop == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_east_stop) (gs);
	return rc;
}


int gtc_scope_slew_west_start (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_west_start == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_west_start) (gs);
	return rc;
}


int gtc_scope_slew_west_stop (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_west_stop == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_west_stop) (gs);
	return rc;
}


int gtc_scope_slew_north_start (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_north_start == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_north_start) (gs);
	return rc;
}


int gtc_scope_slew_north_stop (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_north_stop == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_north_stop) (gs);
	return rc;
}


int gtc_scope_slew_south_start (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_south_start == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_south_start) (gs);
	return rc;
}


int gtc_scope_slew_south_stop (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_south_stop == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_south_stop) (gs);
	return rc;
}


int gtc_scope_slew_stop (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_slew_stop == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_slew_stop) (gs);
	return rc;
}



int gtc_scope_get_ra_slewrate (GtcScope *gs, double *degsec)
{
	int rc;
	if (gs == NULL || degsec == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_ra_slewrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_ra_slewrate) (gs, degsec);
	return rc;
}


int gtc_scope_set_ra_slewrate (GtcScope *gs, double degsec) 
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_ra_slewrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_ra_slewrate) (gs, degsec);
	return rc;
}


int gtc_scope_get_dec_slewrate (GtcScope *gs, double *degsec)
{
	int rc;
	if (gs == NULL || degsec == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_dec_slewrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_dec_slewrate) (gs, degsec);
	return rc;
}


int gtc_scope_set_dec_slewrate (GtcScope *gs, double degsec) 
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_dec_slewrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_dec_slewrate) (gs, degsec);
	return rc;
}



int gtc_scope_get_trackrate (GtcScope *gs, double *hertz)
{
	int rc;
	if (gs == NULL || hertz == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_trackrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_trackrate) (gs, hertz);
	return rc;
}


int gtc_scope_set_trackrate (GtcScope *gs, double hertz)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_trackrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_trackrate) (gs, hertz);
	return rc;
}



int gtc_scope_get_guiderate (GtcScope *gs, double *arcsecsec)
{
	int rc;
	if (gs == NULL || arcsecsec == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_guiderate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_guiderate) (gs, arcsecsec);
	return rc;
}


int gtc_scope_set_guiderate (GtcScope *gs, double arcsecsec)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_guiderate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_guiderate) (gs, arcsecsec);
	return rc;
}



/* scope focus */
int gtc_scope_focus_in_start (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_focus_in_start == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_focus_in_start) (gs);
	return rc;
}


int gtc_scope_focus_out_start (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_focus_out_start == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_focus_out_start) (gs);
	return rc;
}


int gtc_scope_focus_stop (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_focus_stop == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_focus_stop) (gs);
	return rc;
}


int gtc_scope_focus_get_speed (GtcScope *gs, double *speed)
{
	int rc;
	if (gs == NULL || speed == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_focus_get_speed == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_focus_get_speed) (gs, speed);
	return rc;
}


int gtc_scope_focus_set_speed (GtcScope *gs, double speed)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_focus_set_speed == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_focus_set_speed) (gs, speed);
	return rc;
}



/* scope reticle */
int gtc_scope_reticle_get_brightness (GtcScope *gs, double *brightness)
{
	int rc;
	if (gs == NULL || brightness == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_reticle_get_brightness == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_reticle_get_brightness) (gs,
								brightness);
	return rc;
}


int gtc_scope_reticle_set_brightness (GtcScope *gs, double brightness)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_reticle_set_brightness == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_reticle_set_brightness) (gs,
								brightness);
	return rc;
}


int gtc_scope_reticle_get_flashrate (GtcScope *gs, double *rate)
{
	int rc;
	if (gs == NULL || rate == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_reticle_get_flashrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_reticle_get_flashrate) (gs,
								rate);
	return rc;
}


int gtc_scope_reticle_set_flashrate (GtcScope *gs, double rate)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_reticle_set_flashrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_reticle_set_flashrate) (gs,
								rate);
	return rc;
}


int gtc_scope_reticle_get_dutyrate (GtcScope *gs, double *rate)
{
	int rc;
	if (gs == NULL || rate == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_reticle_get_dutyrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_reticle_get_dutyrate) (gs,
								rate);
	return rc;
}


int gtc_scope_reticle_set_dutyrate (GtcScope *gs, double rate)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_reticle_set_dutyrate == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_reticle_set_dutyrate) (gs,
								rate);
	return rc;
}



/* scope field derotator */
int gtc_scope_derotator_enable (GtcScope *gs, int enable)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_derotator_enable == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_derotator_enable) (gs, enable);
	return rc;
}



/* scope targeting/tracking */
int gtc_scope_get_current_ra (GtcScope *gs, GtcRa *ra)
{
	int rc;
	if (gs == NULL || ra == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_current_ra == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_current_ra) (gs, ra);
	return rc;
}


int gtc_scope_get_current_dec (GtcScope *gs, GtcDec *dec)
{
	int rc;
	if (gs == NULL || dec == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_current_dec == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_current_dec) (gs, dec);
	return rc;
}


int gtc_scope_get_target_ra (GtcScope *gs, GtcRa *ra)
{
	int rc;
	if (gs == NULL || ra == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_target_ra == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_target_ra) (gs, ra);
	return rc;
}


int gtc_scope_set_target_ra (GtcScope *gs, GtcRa *ra)
{
	int rc;
	if (gs == NULL || ra == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_target_ra == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_target_ra) (gs, ra);
	return rc;
}


int gtc_scope_get_target_dec (GtcScope *gs, GtcDec *dec)
{
	int rc;
	if (gs == NULL || dec == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_get_target_dec == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_get_target_dec) (gs, dec);
	return rc;
}


int gtc_scope_set_target_dec (GtcScope *gs, GtcDec *dec)
{
	int rc;
	if (gs == NULL || dec == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_set_target_dec == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_set_target_dec) (gs, dec);
	return rc;
}


int gtc_scope_goto_target (GtcScope *gs)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_goto_target == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_goto_target) (gs);
	return rc;
}



/* scope library objects */
int gtc_scope_object_set_target (GtcScope *gs, int catalog, long object)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_object_set_target == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_object_set_target) (gs,
							catalog,
							object);
	return rc;
}



/* scope Periodic Error Correction (PEC */
int gtc_scope_pec_ra_enable (GtcScope *gs, int enable)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_pec_ra_enable == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_pec_ra_enable) (gs, enable);
	return rc;
}


int gtc_scope_pec_dec_enable (GtcScope *gs, int enable)
{
	int rc;
	if (gs == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_pec_dec_enable == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_pec_dec_enable) (gs, enable);
	return rc;
}



/* scope options */
int gtc_scope_option_list (GtcScope *gs, GtcOption **optionlist)
{
	int rc;
	if (gs == NULL || optionlist == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_list == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_list) (gs, optionlist);
	return rc;
}


int gtc_scope_option_get_int (GtcScope *gs, char *name, int *value)
{
	int rc;
	if (gs == NULL || name == NULL || value == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_get_int == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_get_int) (gs,
							name,
							value);
	return rc;
}


int gtc_scope_option_set_int (GtcScope *gs, char *name, int value)
{
	int rc;
	if (gs == NULL || name == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_set_int == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_set_int) (gs,
							name,
							value);
	return rc;
}


int gtc_scope_option_get_boolean (GtcScope *gs, char *name, int *value)
{
	int rc;
	if (gs == NULL || name == NULL || value == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_get_boolean == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_get_boolean) (gs,
							name,
							value);
	return rc;
}


int gtc_scope_option_set_boolean (GtcScope *gs, char *name, int value)
{
	int rc;
	if (gs == NULL || name == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_set_boolean == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_set_boolean) (gs,
							name,
							value);
	return rc;
}


int gtc_scope_option_get_double (GtcScope *gs, char *name, double *value)
{
	int rc;
	if (gs == NULL || name == NULL || value == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_set_double == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_get_double) (gs,
							name,
							value);
	return rc;
}


int gtc_scope_option_set_double (GtcScope *gs, char *name, double value)
{
	int rc;
	if (gs == NULL || name == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_set_double == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_set_double) (gs,
							name,
							value);
	return rc;
}


int gtc_scope_option_get_string (GtcScope *gs, char *name, char **value)
{
	int rc;
	if (gs == NULL || name == NULL || value == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_get_string == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_get_string) (gs,
							name,
							value);
	return rc;
}


int gtc_scope_option_set_string (GtcScope *gs, char *name, char *value)
{
	int rc;
	if (gs == NULL || name == NULL || value == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_option_set_string == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_option_set_string) (gs,
							name,
							value);
	return rc;
}



/* observatory control elements */
int gtc_observatory_new (GtcObservatory **gop, GtcModule *gm)
{
	GtcObservatory *go = NULL;
	if (gop == NULL || gm == NULL)
		return GTC_RC_INVALIDARGUMENT;
	*gop = NULL;
	go = calloc (1, sizeof (GtcObservatory));
	if (go == NULL)
		return GTC_RC_NOMEMORY;
	go->module = gm;
	gm->refcount++;
	*gop = go;
	return 0;
}


int gtc_observatory_connect (GtcObservatory *go, char *device)
{
	int rc;
	if (go == NULL || device == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (go->flags & GTC_FLAG_DEVICECONNECTED)
		return GTC_RC_DEVICEBUSY;
	if (go->module->moduleinfo->vector->observatory_connect == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_connect) (go, device);
	if (rc == 0) {
		go->flags |= GTC_FLAG_DEVICECONNECTED;
	}
	return rc;
}


int gtc_observatory_close (GtcObservatory *go)
{
	int rc;
	if (go == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(go->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (go->module->moduleinfo->vector->observatory_close == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_close) (go);
	if (rc == 0) {
		go->flags -= (go->flags & GTC_FLAG_DEVICECONNECTED);
	}
	return rc;
}


int gtc_observatory_free (GtcObservatory **gop)
{
	GtcObservatory *go = NULL;
	if (gop == NULL || *gop == NULL)
		return GTC_RC_INVALIDARGUMENT;
	go = *gop;
	go->module->refcount--;
	free (go);
	*gop = NULL;
	return 0;
}


int gtc_observatory_slew_cw_start (GtcObservatory *go)
{
	int rc;
	if (go == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(go->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (go->module->moduleinfo->vector->observatory_slew_cw_start == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_slew_cw_start) (go);
	return rc;
}


int gtc_observatory_slew_ccw_start (GtcObservatory *go)
{
	int rc;
	if (go == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(go->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (go->module->moduleinfo->vector->observatory_slew_ccw_start == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_slew_ccw_start) (go);
	return rc;
}


int gtc_observatory_slew_stop (GtcObservatory *go)
{
	int rc;
	if (go == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(go->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (go->module->moduleinfo->vector->observatory_slew_stop == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_slew_stop) (go);
	return rc;
}


int gtc_observatory_skylight_open (GtcObservatory *go)
{
	int rc;
	if (go == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(go->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (go->module->moduleinfo->vector->observatory_skylight_open == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_skylight_open) (go);
	return rc;
}


int gtc_observatory_skylight_close (GtcObservatory *go)
{
	int rc;
	if (go == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(go->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (go->module->moduleinfo->vector->observatory_skylight_close == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_skylight_close) (go);
	return rc;
}


int gtc_observatory_light_enable (GtcObservatory *go, int enable)
{
	int rc;
	if (go == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(go->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (go->module->moduleinfo->vector->observatory_light_enable == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_light_enable) (go,
								enable);
	return rc;
}


int gtc_observatory_nightlight_enable (GtcObservatory *go, int enable)
{
	int rc;
	if (go == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(go->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (go->module->moduleinfo->vector->observatory_nightlight_enable == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*go->module->moduleinfo->vector->observatory_nightlight_enable) (go,
								enable);
	return rc;
}


/* Retreive telescope control values */
int gtc_scope_control_get_slewrates (GtcScope *gs, GtcControl **gcp)
{
	int rc;
	if (gs == NULL || gcp == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_control_get_slewrates == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_control_get_slewrates) (gs,
								gcp);
	return rc;
}


int gtc_scope_control_get_trackrates (GtcScope *gs, GtcControl **gcp)
{
	int rc;
	if (gs == NULL || gcp == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_control_get_trackrates == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_control_get_trackrates) (gs,
								gcp);
	return rc;
}


int gtc_scope_control_get_guiderates (GtcScope *gs, GtcControl **gcp)
{
	int rc;
	if (gs == NULL || gcp == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_control_get_guiderates == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_control_get_guiderates) (gs,
								gcp);
	return rc;
}


int gtc_scope_control_get_focus_speeds (GtcScope *gs, GtcControl **gcp)
{
	int rc;
	if (gs == NULL || gcp == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_control_get_focus_speeds == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_control_get_focus_speeds) (gs,
								gcp);
	return rc;
}


int gtc_scope_control_get_reticle_flashrates (GtcScope *gs, GtcControl **gcp)
{
	int rc;
	if (gs == NULL || gcp == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_control_get_reticle_flashrates == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_control_get_reticle_flashrates) (gs,
								gcp);
	return rc;
}


int gtc_scope_control_get_reticle_dutyrates (GtcScope *gs, GtcControl **gcp)
{
	int rc;
	if (gs == NULL || gcp == NULL)
		return GTC_RC_INVALIDARGUMENT;
	if (!(gs->flags & GTC_FLAG_DEVICECONNECTED))
		return GTC_RC_DEVICENOTPRESENT;
	if (gs->module->moduleinfo->vector->scope_control_get_reticle_dutyrates == NULL)
		return GTC_RC_UNIMPLEMENTED;
	rc = (*gs->module->moduleinfo->vector->scope_control_get_reticle_dutyrates) (gs,
								gcp);
	return rc;
}


/* Error message functions */
const char *gtc_get_strerror (int error)
{
	if (error > 0 && error < 125) {
		/* libc error codes */
		return strerror (error);
	} else if (error > GTC_RC_BASEERROR && error < GTC_RC_MAXERROR) {
		return GtcErrorString[error - GTC_RC_BASEERROR];
	}
	return NULL;
}



